# this is shared configuration space 

"""ray_path contains default path to the ray installation
"""
ray_path = None

"""ray_binary contains name of the default binary or shell script
"""
ray_binary = "rayui.sh"



from importlib.resources import path
import numpy as np
import os
import warnings
from natsort import natsorted, ns

from .rml import RMLFile

class RayProperties(np.ndarray):
    """RayProperties class privides simplified interface to access results of the 
    postprocessing of calculated rays and provides functions to read and write those results
    into a txt file, including the header

    Args:
        input (nd.array): create RayProperties from an array. Intentende for internal use only!
        filename (str): file name to load the data from (including the header information/row)
    """
    def __new__(cls,input:np.ndarray=None,/,filename=None) -> None:
        dt_names = ['SourcePhotonFlux', 'NumberRaysSurvived', 'PercentageRaysSurvived', 'PhotonFlux', 'Bandwidth', 'HorizontalFocusFWHM', 'VerticalFocusFWHM']
        dt_formats = [float for n in dt_names]
        dt = np.dtype({'names':dt_names, 'formats':dt_formats})

        if input is None:
            if filename is None:
                input_array = np.zeros(1, dtype=dt)
            else:
                input_array = np.genfromtxt(filename, dtype=float, delimiter='\t', names=True)
        else:
            input_array = input.copy()
            input_array.dtype = dt

        obj = np.asarray(input_array).view(cls)
        return obj

    def save(self,filename):
        """Save current object into tab-separated file with header row

        Args:
            filename (_type_): _description_
        """
        tmp = self.copy()
        if tmp.shape[0]>1:
            tmp.dtype = float
        np.savetxt(filename,tmp,delimiter='\t',header="\t".join(self.dtype.names))
        

    def concat(self,other):
        """concatenate another instalnce of RayProperties and return a new object 

        Args:
            other (RayProperties): instance of RayProperties to be contated to this instance

        Returns:
            RayProperties: resulting RayProperties object
        """
        return RayProperties(np.vstack((self,other)))


class PostProcess():
    """class to post-process the data. 

    It works only if the exported data are RawRaysOutgoing
    """
    def __init__(self) -> None:
    
        self.format_saved_files = '.dat'
        self.source = None
        pass

    def _list_files(self,dir_path:str, end_filename:str):
        """List all the files in dir_path ending with end_filename

        Args:
            dir_path (str): path to a folder
            end_filename (str): the listed files end with end_filename

        Returns:
            res (list): list of files in dir_path ending eith end_filename
        """        
        # list to store files
        res = []
        for file in os.listdir(dir_path):
            # check only text files
            if file.endswith(end_filename):
                res.append(os.path.join(dir_path,file))
        return natsorted(res, alg=ns.IGNORECASE)

    def _extract_fwhm(self,rays:np.array):
        """Calculate the fwhm of the rays.

        If less than 100 rays are passed check what is the standard deviation of the array.
        Else, make an histogram with 30 bins and check when the array falls at less than half the max

        Args:
            rays (np:array): the energy of the x-rays

        Returns:
            float: fwhm
        """        
        # if I have less than 100 rays calculate the standard deviation
        if rays.shape[0]<100:
            return 2*np.sqrt(2*np.log(2))*np.std(rays)
        
        # else actually look for the fwhm

        # make an histogram, get back a tuple of values and bins
        gh = np.histogram(rays, bins=30)
        y = gh[0]
        x_bins = gh[1]

        # take the average of each pari of bins to get the middle
        x = (x_bins[1:] + x_bins[:-1]) / 2

        # Find the maximum y value
        max_y = np.amax(y)  

        # check where y becomes higher that max_y/2
        xs = [x for x in range(y.shape[0]) if y[x] > max_y/2.0]
        fwhm = x[np.amax(xs)-1]-x[np.amin(xs)-1]
        return fwhm
        

        
        

    def _extract_intensity(self,rays:np.array):
        """calculate how many rays there are

        Args:
            rays (np.array): contains rays information
        """        
        return(rays.shape[0])
    
    def extract_nrays_from_source(self, rml_filename):
        """Extract photon flux from rml file, find source automatically

        Args:
            rml_filename (str): the rml file to use to extract the photon flux

        Returns:
            str: the photon flux
        """        
        s = RMLFile(rml_filename)
        for oe in s.beamline.children():
                if hasattr(oe,"photonFlux"):
                    self.source = oe
                    break
        return self.source.photonFlux.cdata, self.source.numberRays.cdata
    
    def postprocess_RawRays(self,exported_element:str=None, exported_object:str=None, dir_path:str=None, sim_number:str=None, rml_filename:str=None):
        """ PostProcess rountine of the RawRaysOutgoing extracted files.

        The method looks in the folder dir_path for a file with the filename:
        :code:`filename = os.path.join(dir_path,sim_number+exported_element + '-' + exported_object+'.csv')`
        for each file it calculates the number of rays, the bandwidth, and the horizontal and vertical focus size,
        it saves it in an array that is composed by :code:`[n_rays,bandwidth,hor_focus,vert_focus]`, that is then saved to
        :code:`os.path.join(dir_path, sim_number+exported_element+'_analyzed_rays.npy')`
        
        Args:
            exported_element (list, optional): a list of containing the exported elements name as str. Defaults to None.
            exported_object (str, optional): the exported object, tested only with RawRaysOutgoing. Defaults to None.
            dir_path (str, optional): the folder where the file to process is located. Defaults to None.
            sim_number (str, optional): the prefix of the file, that is the simulation number with a _prepended, ie `0_`. Defaults to None.
        """        
        source_photon_flux, source_n_rays  = self.extract_nrays_from_source(rml_filename)
        
        source_photon_flux = float(source_photon_flux)
        source_n_rays      = float(source_n_rays)
        filename = os.path.join(dir_path,sim_number+exported_element + '-' + exported_object+'.csv')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            #rays = np.loadtxt(filename, skiprows=2)
            rays = np.genfromtxt(filename, dtype=float, delimiter='\t', names=True,skip_header=1)
        ray_properties = RayProperties()
        if rays.shape[0]==0: # if no rays survived
            # source photon flux
            ray_properties['SourcePhotonFlux'] = source_photon_flux
            pass
        else:
            # source photon flux
            ray_properties['SourcePhotonFlux'] = source_photon_flux
            # number of rays reaching the oe
            ray_properties['NumberRaysSurvived'] = self._extract_intensity(rays)
            # percentage of survived photons
            ray_properties['PercentageRaysSurvived'] = ray_properties['NumberRaysSurvived']/source_n_rays*100
            # photon flux reaching the oe
            ray_properties['PhotonFlux'] = source_photon_flux/100*ray_properties['PercentageRaysSurvived']
            # bandwidth of the rays reaching the oe
            ray_properties['Bandwidth'] = self._extract_fwhm(rays[f'{exported_element}_EN'])
            # horizontal focus
            ray_properties['HorizontalFocusFWHM'] = self._extract_fwhm(rays[f'{exported_element}_OX'])
            # vertical focus
            ray_properties['VerticalFocusFWHM'] = self._extract_fwhm(rays[f'{exported_element}_OY'])
        
        new_filename = os.path.join(dir_path, sim_number+exported_element+'_analyzed_rays.dat')
        #self._save_file(new_filename, ray_properties)
        ray_properties.save(new_filename)
        return 

    def cleanup(self,dir_path:str=None, repeat:int=1, exp_elements:list=None):
        """Reads all the results of the postprocessing process and summarize 
        them in a single file for each exported object.
        
        This functions reads all the temporary files created by :code:`self.postptocess_RawRays()`
        saves one file for each exported element in dir_path, and deletes the temporary files.
        If more than one round of simulations was done, the values are averaged.

        Args:
            dir_path (str, optional): The path to the folder to cleanup. Defaults to None.
            repeat (int, optional): number of rounds of simulations. Defaults to 1.
            exp_elements (list, optional): the exported elements names as str. Defaults to None.
        """        
        header = "SourcePhotonFlux\t\t NumberRaysSurvived\t\t  PercentageRaysSurvived   PhotonFlux\t\t\t\tBandwidth\t\t\t\t HorizontalFocusFWHM\t  VerticalFocusFWHM"
        for d in exp_elements:
            for r in range(repeat):
                dir_path_round=os.path.join(dir_path,"round_"+str(r))
                files = self._list_files(dir_path_round, d[0]+"_analyzed_rays"+self.format_saved_files)
                for f_ind, f in enumerate(files):
                    if r == 0 and f_ind==0:
                        analyzed_rays = RayProperties(filename=f)
                    elif r==0 and f_ind!=0:
                        tmp = RayProperties(filename=f)
                        analyzed_rays = analyzed_rays.concat(tmp)
                    elif r>=1:
                        tmp = RayProperties(filename=f)
                        for n in analyzed_rays.dtype.names: analyzed_rays[n] += tmp[n]
                    else:
                        pass
            fn = os.path.join(dir_path, d[0])
            for n in analyzed_rays.dtype.names: analyzed_rays[n] /= repeat
            analyzed_rays.save(f"{fn}.dat")

class PostProcessAnalyzed():
    """class to analyze the data exported by RAY-UI
    """
    def __init__(self) -> None:
        pass

    def retrieve_flux_beamline(self, folder_name,source,oe,nsimulations,rounds=1,current=0.3):
        """Extract the flux from object ScalarBeamProperties and from source ScalarElementProperties.

        This function takes as arguments the name of the 
        simulation folder, the exported objet in RAY-UI and there
        number of simulations and returns the flux at the optical element in 
        percentage and in number of photons, and the flux produced
        by the dipole.
        It requires ScalarBeamProperties to be exported for the desired optical element,
        if the source is a dipole it requires ScalarElementProperties to be exported for the Dipole

        Args:
            folder_name (str): the path to the folder where the simulations are
            source (str): the source name
            oe (str): the optical element name
            nsimulations (int): the number of simulations
            rounds (int): the number of rounds of simulations
            current (float, optional): the ring current in Ampere. Defaults to 0.3.

        Returns:
            if the source is a Dipole:
                photon_flux (np.array) : the photon flux at the optical element
                flux_percent (np.array) : the photon flux in percentage relative to the source
                source_Photon_flux (np.array) : the photon flux of the source
            else:
                flux_percent (np.array) : the photon flux in percentage relative to the source
        """        
        scale_factor = current/0.1
        flux_percent = np.zeros(nsimulations)
        if source == 'Dipole':
            flux         = np.zeros(nsimulations)
            flux_dipole  = np.zeros(nsimulations)
        for n in range(nsimulations):
            try:
                for r in range(rounds):
                    temp = np.loadtxt(folder_name+'/round_'+str(r)+'/'+str(n)+'_'+oe+'-ScalarBeamProperties.csv',skiprows = 2)
                    flux_percent[n] += temp[25]/rounds
                    if source == 'Dipole':
                        dipole_abs_flux = np.loadtxt(folder_name+'/round_'+str(r)+'/'+str(n)+'_Dipole-ScalarElementProperties.csv',skiprows = 2)
                        flux[n]         += (dipole_abs_flux[12]*temp[25]/100)/rounds
                        flux_dipole[n]  += dipole_abs_flux[12]/rounds
                
            except OSError:
                print('######################')
                print(n, 'NOT FOUND:\n'+folder_name+'/round_'+str(r)+'/'+str(n)+'_'+oe+'-ScalarBeamProperties.csv')
                continue
        flux_percent = np.array(flux_percent)
        if source == 'Dipole':
            flux = np.array(flux)
            flux_dipole = np.array(flux_dipole)
            return flux*scale_factor, flux_percent, flux_dipole*scale_factor
        return flux_percent*scale_factor

    def retrieve_bw_and_focusSize(self,folder_name:str,oe:str,nsimulations:int,rounds:int):
        """Extract the bandwidth and focus size from ScalarBeamProperties of an object.

        Args:
            folder_name (str): the path to the folder where the simulations are 
            oe (str): the optical element name
            nsimulations (int): the number of simulations
            rounds (int): the number of rounds of simulations

        Returns:
            bw np.array: the bandwidth
            foc_x np.array: the horizontal focus
            foc_y np.array: the vertical focus
        """        
        bw_ind        =10
        fx_ind        =4
        fy_ind        =7
        bw            = np.zeros(nsimulations)
        foc_x         = np.zeros(nsimulations)
        foc_y         = np.zeros(nsimulations)
        n0=0
        for j in range(rounds):
            for n in range(nsimulations):
                try:
                    temp = np.loadtxt(folder_name+'/round_'+str(j)+
                                    '/'+str(n)+'_'+oe+'-ScalarBeamProperties.csv',
                                    skiprows = 2)
                    bw[n]     += (temp[bw_ind])/rounds
                    foc_x[n]  += (temp[fx_ind])/rounds
                    foc_y[n]  += (temp[fy_ind])/rounds
                    if n0==(nsimulations-1):
                        n0=0
                    else:
                        n0+=1
                    
                except OSError:
                    print('######################')
                    print('NOT FOUND:\n'+folder_name+'/round_'+str(j)+
                                    '/'+str(n)+'_'+oe+'-ScalarBeamProperties.csv')
                    bw[n]     += 0
                    foc_x[n]  += 0
                    foc_y[n]  += 0
        return bw,foc_x,foc_y

    def moving_average(self, x, w):
        """Computes the morivng average with window w on the array x

        Args:
            x (array): the array to average
            w (int): the window for the moving average

        Returns:
            np.array: the x array once the moving average was applied
        """        
        if w == 0:
            return x
        return np.convolve(x, np.ones(w), 'valid') / w





import typing

from .dictionaries import *
from .attributes import *


###############################################################################
class XmlElement:
    """Representation of an XML element in its raw form

    Raises:
        AttributeError: _description_
    """
    #####################################
    def __init__(self, name:str, attributes:typing.MutableMapping, parent=None, **kwargs):
        """_summary_

        Args:
            name (str): element name
            attributes (dict): element attributes as a dict-like object
        """
        if name is not None and not isinstance(name,str):
            raise TypeError("name parameter of 'XmlElement' class must be a string or 'None', got {}".format(type(name)))
        self._original_name = name
        self._name = sanitizeName(name)
        if attributes is not None and not isinstance(attributes,typing.MutableMapping):
            raise TypeError("attributes parameter of 'XmlElement' class must be a dict-line object or 'None', got {}".format(type(attributes)))
        
        # converting attributes ...
        _attributes = SafeValueDict()
        if attributes is not None:
            if isinstance(attributes,MappedDict):
                _attribute_items = attributes.original().items()
            else:
                _attribute_items = attributes.items()
            for k,v in _attribute_items:
                _attributes[k] = v#XmlAttribute(v)
        self._attributes = _attributes

        self._children = []
        self.is_root = False
        self.cdata = ""
        self._parent = parent

    def get_full_path(self):
        """Returns the full path of the xml object

        Returns:
            str: path of the xml object
        """        
        if self._parent is None:
            return self.resolvable_name()
        else:
            return self._parent.get_full_path()+"."+self.resolvable_name()

    def resolvable_name(self):
        """Returns the name of the objects, removing lab.beamline.

        Returns:
            str: name of the object
        """        
        if self._parent is None:
            return self._name
        else:
            if hasattr(self._parent,"_name_attribute"):
                return self[self._parent._name_attribute]
            else:
                return self._name#"NOTFOUND"

    #@property
    def children(self):
        return self._children

    #@property
    def attributes(self):
        return self._attributes

    #@property
    def original_name(self):
        return self._original_name

    #@property
    def name(self):
        return self._name

    def add_child(self, element):
        """
        Store child elements.
        """
        self._children.append(element)

    def add_cdata(self, cdata):
        """
        Store cdata
        """
        self.cdata = self.cdata + cdata

    def get_attribute(self, key):
        """
        Get attributes by key
        """
        return self._attributes.get(key)

    def get_elements(self, name=None):
        """
        Find a child element by name
        """
        if name:
            return [e for e in self._children if e._name == name]
        else:
            return self._children


    # dictionary line access to the elements
    def __getitem__(self, key):
        return self.get_attribute(key)

    # attribute aka ./dot access to the fields
    def __getattr__(self, key):
        matching_children = [x for x in self._children if x._name == key]
        if matching_children:
            if len(matching_children) == 1:
                self.__dict__[key] = matching_children[0]
                return matching_children[0]
            else:
                self.__dict__[key] = matching_children
                return matching_children
        else:
            raise AttributeError("'%s' has no attribute '%s'" % (self._name, key))

    def __hasattribute__(self, name):
        if name in self.__dict__:
            return True
        return any(x._name == name for x in self._children)

    def __iter__(self):
        yield self

    def __str__(self):
        return "XmlElement <%s> with attributes %s, children %s and cdata %s" % (
            self._name,
            self._attributes,
            self._children,
            self.cdata,
        )

    def __repr__(self):
        if True:
            return "XmlElement(name = %s, attributes = %s, cdata = %s)" % (
                self._name,
                self._attributes,
                self.cdata,
            )
        else:
            return f'{self.id}'

    def __nonzero__(self):
        return self.is_root or self._name is not None

    def __eq__(self, val):
        return self.cdata == val


    def __hash__(self) -> int:
        return hash((tuple(self._children),tuple(self._attributes),tuple(self._name)))

    def __dir__(self):
        children_names = [x._name for x in self._children]
        return children_names

    def __len__(self):
        return len(self._children)

    def __contains__(self, key):
        return key in dir(self)


    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass#return True

###############################################################################
class XmlAttributedNameElement(XmlElement):
    def __init__(self, name_attribute:str, name: str, attributes: dict, **kwargs):
        super().__init__(name, attributes,**kwargs)
        self._name_attribute = name_attribute


    def __dir__(self):
        """enumerating child objects by its attibute name

        Returns:
            _type_: _description_
        """
        children_names = [x._attributes[self._name_attribute] for x in self._children]
        return children_names

    # def __setattr__(self, __name: str, __value) -> None:
    #     if __name!="_name" and __name!="_attributes" and __name!="children" and __name!="is_root"  and __name!="cdata" and __name!="_name_attribute":
    #         raise AttributeError("XmlAttributedNameElement object attribute '{}' is read-only".format(__name))

    # def __setattr__(self, __name: str, __value) -> None:
    #     if hasattr(self,__name):
    #         return super().__setattr__(__name, __value)
    #     else:
    #         raise AttributeError("XmlAttributedNameElement object attribute '{}' is read-only".format(__name))

    def __getattr__(self, key):
        matching_children = [x for x in self._children if x._attributes[self._name_attribute] == key]
        if matching_children:
            if len(matching_children) == 1:
                self.__dict__[key] = matching_children[0]
                return matching_children[0]
            else:
                self.__dict__[key] = matching_children
                return matching_children
        else:
            raise AttributeError("'%s' has no attribute '%s'" % (self._name, key))


###############################################################################
# RayUI process handling and RayUI command API

from sre_constants import SUCCESS
from . import config
from .errors import RayPyRunnerError,RayPyError,TimeoutError
import os
import subprocess,signal
import time
import psutil

#import sys # used for some debugging

###############################################################################
class RayUIRunner:
    """RayUIRunner class implements all logic to start a RayUI process, load and rml file, trace and export.

    Args:
            ray_path (str, optional): the path to the RAY-UI installation folder. 
                                      Defaults to config.ray_path, that will look for the
                                      ray_path in the standard installation folders.
            ray_binary (_type_, optional): the binary file of RAY-UI. 
                                            Defaults to "rayui.sh".
            background (bool, optional): activate background mode. Defaults to True.
            hide (bool, optional): Hide the RAY-UI graphical instances. 
                                   Available only if xvfb is installed. 
                                   Defaults to False.
    """
    def __init__(self,ray_path=config.ray_path,ray_binary=config.ray_binary,background=True,hide=False) -> None:
        """
        Args:
            ray_path (str, optional): the path to the RAY-UI installation folder. 
                                      Defaults to config.ray_path, that will look for the
                                      ray_path in the standard installation folders.
            ray_binary (_type_, optional): the binary file of RAY-UI. 
                                            Defaults to "rayui.sh".
            background (bool, optional): activate background mode. Defaults to True.
            hide (bool, optional): Hide the RAY-UI graphical instances. 
                                   Available only if xvfb is installed. 
                                   Defaults to False.

        Raises:
            Exception: _description_
        """        
        if ray_path is None:
            ray_path = self.__detect_ray_path()
        if ray_path is None:
            raise Exception("ray_path must be defined for now!")
        self._path = ray_path
        self._binary = ray_binary
        self._options = "-b" if background else ""
        self._process = None
        self._verbose = False
        if hide:
            self._hide = "xvfb-run --auto-servernum --server-num=1 "
        else: 
            self._hide = ''



        # internal configuration parameters
        self._auto_flush = True     # flush on write calls

    def run(self):
        """Open one instance of RAY-UI using subprocess

        Raises:
            RayPyRunnerError: if the RAY-UI executable is not found raise an error

        """        
        if not self.isrunning:
            fullpath = os.path.join(self._path,self._binary)
            if not os.path.isfile(fullpath):
                raise RayPyRunnerError("Ray executable {0} is not found".format(fullpath))
            fullpath = self._hide + fullpath
            env = dict(os.environ) # TODO:: rethink a bit about this line 
            self._process = subprocess.Popen(
                                            fullpath+" -b", 
                                            shell=True, 
                                            stdin=subprocess.PIPE, 
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.STDOUT, 
                                            env=env
                                            )
        return self

    @property
    def isrunning(self):
        """Check weather a process is running and rerutn a boolean

        Returns:
            bool: returns True if the process is running, otherwise False
        """        
        if self._process is None:
            return False
        else:
            poll_result = self._process.poll()
            if poll_result is not None:
                # we have an exit code, so process is not valid any more and clean it it up
                self._process = None
                return False
            else:
                return True

    def kill(self):
        """kill a RAY-UI process
        """
        if self.isrunning:
            pid = self._process.pid
            process = psutil.Process(pid)
            for proc in process.children(recursive=True):
                proc.kill()
            process.kill()

    @property
    def pid(self):
        """Get process id of the RayUI process

        Returns:
            int: PID of the process if it running, None otherwise
        """
        if self.isrunning:
            return self._process.pid
        else:
            return None

    def _write(self, instr:str, endline="\n"):
        """Write command to RayUI interface

        Args:
            instr (str): _description_
            endline (str, optional): _description_. Defaults to endline character.

        Raises:
            RayPyRunnerError: _description_
        """     
        if self.isrunning:
            payload = bytes(instr+endline,"utf8")
            self._process.stdin.write(payload)
            if self._auto_flush:
                self._process.stdin.flush()
        else:
            raise RayPyRunnerError("RayUI process is not started")
    
    def _readline(self)->str:
        """Read a line from the stdout of the process and convert to a string

        Returns:
            str: line read from the input
        """
        if self.isrunning:
            line =  self._process.stdout.readline().decode('utf8').rstrip('\n')
            if self._verbose: # verbose
                print(line)
            return line
        else:
            return None


    def __detect_ray_path(self) -> str:
        """Internal function to autodetect installation path of RAY-UI

        Raises:
            RayPyRunnerError: is case no RAY-UI installations can be detected

        Returns:
            str: string with the detected RAY-UI installation path
        """
        basepaths = ("~", "~/Applications","/opt","/Applications")
        installpaths = ("RAY-UI-development","RAY-UI")
        pathlist = [os.path.expanduser(p) for p in [os.path.join(x,y) for x in basepaths for y in installpaths]]

        for ray_path in pathlist:
            if os.path.isdir(ray_path):
                return ray_path
        raise RayPyRunnerError("Can not detect rayui installation path!")
 
###############################################################################
class RayUIAPI:
    """RayUIAPI class implements (hopefully all) command interface of the RAY-UI
    """
    def __init__(self,runner:RayUIRunner=None) -> None:
        """Optional Reference to an existing RayUIRunner
        Args:
            runner (RayUIRunner, optional): reference to existing runner. 
                                            If set to None a new runner instance will 
                                            be automaticlly created. Defaults to None.
        """
        if runner is None:
            runner = RayUIRunner().run()
        self._runner = runner
        self._read_wait_delay = 0.01    # if rayui does not send anything to stdio this delay will be used before next attempt to read
        self._quit_timeout = 300         # default timeout for commands like quit

    def quit(self):
        """quit RAY-UI if it is running
        """
        if self._runner.isrunning:
            self._runner._write("quit")
            try:
                self._runner._process.wait(self._quit_timeout)
            except subprocess.TimeoutExpired:
                raise TimeoutError("Timeout while trying to quit")
        
    def load(self,rml_path,**kwargs):
        """Load an rml file

        Args:
            rml_path (str): path to the rml file
        """        
        return self._cmd_io("load",rml_path,**kwargs)

    def save(self,rml_path,**kwargs):
        """Save an rml file

        Args:
            rml_path (path): path to save the rml file
        """        
        return self._cmd_io("save",rml_path,**kwargs)

    def trace(self,analyze=True,**kwargs):
        """Trace an rml file (must have been loaded before). 

        Args:
            analyze (bool, optional): If True RAY-UI will perform analysis of the rays. 
                                      Defaults to True.

        """        
        return self._cmd_io("trace",None if analyze else "noanalyze", **kwargs)

    def export(self,objects:str, parameters:str, export_path:str, data_prefix:str, **kwargs):
        """Export simulation results from RAY-UI.

        Args:
            objects (str): string with objects list, e.g. "Dipole,DetectorAtFocus"
            parameters (str): stromg with parameters to export, e.g. "ScalarBeamProperties,ScalarElementProperties"
            export_path (str): path where to save the data
            data_prefix (str): prefix for the putput files
        """
        payload = '"'+objects + '"' + " " + parameters + " " + export_path + " " + data_prefix
        return self._cmd_io("export",payload,**kwargs)

    def _cmd_io(self,cmd:str,payload:str=None,/, cbNewLine=None):
        """The _cmd_io is an internal method which helps to execute a RAY-UI command.
        All commands are run in the following way:
        1. A command is sent to RAY-UI
        2. If Ray-UI acknowleges the command it prints it back
        3. Some extra output can happen (depending on the command)
        4. RAY-UI  writes info "success" or "failed"

        Args:
            cmd (str): string with command (e.g. "load" or "trace")
            payload (str, optional): possible payload for the command (e.g. rml 
                                    file path for the load command). Defaults to None.
        Raises:
            RayPyError: in case of an unsupported reply 

        Returns:
            bool: True on success, False on RAY-UI side error
        """
        if payload is None:
            payload = ""
        cmdstr = cmd+" "+payload
        self._runner._write(cmdstr)
        status = self._wait_for_cmd_io(cmd,cbdataread=cbNewLine)
        if status=="success":# or 'trace success':
            return True
        elif status=="failed":
            return False
        elif status=="ed failed": # specical workaround case for the "loaded failed" reply to laod command
            return False
        else:
            raise RayPyError("Got unsupported reply from ray while waiting for command IO")
        
    def _wait_for_cmd_io(self,cmd,timeout=None,cbdataread=None):
        timecnt = 0.0
        line = ""
        cmd_seen = False # we shall see cmd twice - once as ACK for the execution and once as status return
        while True:
            line = self._runner._readline()
            if line is None:
                time.sleep(self._read_wait_delay)
                timecnt+=self._read_wait_delay
                continue
            if (line.startswith(cmd)):
                if cmd_seen:
                    break
                else:
                    cmd_seen = True
            else:
                if cbdataread is not None:
                    cbdataread(line)
            if timeout is not None and timecnt>timeout:
                raise TimeoutError("timeout while waiting ray command io")
        return line.lstrip(cmd).strip()





from .simulate import Simulate
from .rml import ObjectElement
from .recipes import BeamWaist

import numpy as np
import matplotlib.pyplot as plt
import time
import os
import sys
    

class PlotBeamwaist():
    '''
    To use this class one needs to trace and export RawRaysOutgoing for each optical element (no image planes)
    '''
    def __init__(self, directory:str, sim:Simulate):
        self.directory = 'RAYPy_Simulation_'+directory
        self._original_directory = directory
        if isinstance(sim,Simulate):
            self._sim = sim
        self.lim              = False
        self.step             = False
        self.z                = False
        self.step_z           = False
        self.rot              = False
        self.factor           = False
        self.previous_results = False
        self.count_el         = 0
        self.count_fig        = 0
    
    def simulate_beamline(self, energy:float,/,source:ObjectElement=None,nrays:int=None, force:bool=False):
        #sim.params,sim.exports, sim.simulation_name = ResolvingPower(energy_range, elisa.DetectorAtFocus,ES,cff)
        rp = BeamWaist(energy,source=source,nrays=nrays,sim_folder=self._original_directory, force=force)

        # test resolving power simulations
        self._sim.run(rp, multiprocessing=5, force=force)
        #self._sim.beamwaist_simulation(energy,source=source,nrays=nrays,sim_folder=self._original_directory, force=force)
    
    def _parse_beamline_elements(self, debug=False):
        self.element_names_list = []
        self.distance_list=[]
        self.rotation_list=[]
        for ind, oe in enumerate(self._sim.rml.beamline._children):
                for par in oe:
                    try:
                        # append to distances
                        if ind != 0:
                            self.distance_list.append(float(par.worldPosition.z.cdata)-sum(self.distance_list))                            
                        else:    
                            self.distance_list.append(float(par.worldPosition.z.cdata))
                        # append to rotation
                        if ind == 0:
                            self.rotation_list.append(False)
                        elif hasattr(par, "azimuthalAngle"):
                            if par.azimuthalAngle.cdata == '0' or par.azimuthalAngle.cdata=='180':
                                self.rotation_list.append(False)
                            elif par.azimuthalAngle.cdata == '90' or par.azimuthalAngle.cdata=='270':
                                self.rotation_list.append(True)
                            else: 
                                raise ValueError('Only beamline elements with azimuthal angle 0,90,180,270 are supported', oe.name, par.azimuthalAngle.cdata)
                        else:
                            self.rotation_list.append(False)
                        # append to element names list, only if it is not an imageplane
                        #if oe.get_attribute('type') != 'ImagePlaneBundle' and oe.get_attribute('type') != 'ImagePlane':
                        self.element_names_list.append(oe.attributes().original()['name'])
                    except AttributeError:
                        print(oe.attributes().original()['name'], par.worldPosition.z.cdata)
                        pass
        if debug:
            print('DEBUG:: element_names_list', self.element_names_list)
            print('DEBUG:: distance_list', self.distance_list)
            print('DEBUG:: rotation_list', self.rotation_list)


    def trace_beamwaist(self, save_results:bool=True, element_names_list:list=None):
        self._parse_beamline_elements()
        # I overwrite the names for the moment, since I can not access them automatically
        if element_names_list != None:
            self.element_names_list=element_names_list
        for ind in range(len(self.element_names_list)):
            if ind+1 != len(self.element_names_list):
                self.add_element(name=self.element_names_list[ind],
                                z=self.distance_list[ind+1],
                                rot=self.rotation_list[ind])
        if save_results == True:
            self.save_results()
    
    def reduce_Nrays(self,factor):
        self.factor = factor
        
    def load_previous_results(self, element_names_list=None):
        self._parse_beamline_elements()
        if element_names_list != None:
            self.element_names_list=element_names_list
        start1  = time.time()
        print('Load results...')
        self.xh=np.loadtxt(os.path.join(self.directory,'xh.txt'))
        self.yh=np.loadtxt(os.path.join(self.directory,'yh.txt'))
        stop = time.time()
        print('time:', np.round(stop-start1,2),'s')
        
    def save_results(self):
        np.savetxt(os.path.join(self.directory, 'xh.txt'), self.xh)
        np.savetxt(os.path.join(self.directory, 'yh.txt'), self.yh)
            
    def add_element(self,name,z,rot=False):
        start1  = time.time()
        z       = np.arange(0,z+self.step_z,self.step_z)

        if self.previous_results == False:
            print('Tracing '+name)
            
            rays    = np.loadtxt(os.path.join(self.directory,'round_0', 
                                         '0_'+name
                                         +'-RawRaysOutgoing.csv'), 
                            skiprows=2)
            if self.factor != False:
                max_n_rays = int(rays.shape[0]/self.factor)
            if max_n_rays < 100:
                sys.exit('Set a lower reduction factor, there are no more rays to plot!')
            rays = rays[0:max_n_rays]
            
            
            if self.count_el == 0:
                self.xh,self.yh   = self.trace(z,rays,rot)
                #print("DEBUG:: self.xh.shape, self.yh.shape:",self.xh.shape, self.yh.shape)
            else:
                txh,tyh   = self.trace(z,rays,rot, name)
            if rot == True:
                txh=np.rot90(txh)
                tyh=np.rot90(tyh)
            if self.count_el != 0:
                #print("DEBUG:: txh.shape, tyh.shape:",txh.shape, tyh.shape)
                self.xh=np.concatenate((self.xh,txh),axis=1)
                self.yh=np.concatenate((self.yh,tyh),axis=1)

            stop = time.time()
            print('time:', np.round(stop-start1,2),'s')
        self.count_el += 1
        
    def define_hist(self,lim,step):
        self.lim  = lim
        self.step = step
    
    def define_zstep(self,step_z):
        self.step_z = step_z
    
        
    def trace_pos(self,rays,z):
        ox     = 3
        oy     = 4
        oz     = 5
        dx     = 6
        dy     = 7
        dz     = 8
        
        dxz = rays[:,dx]/rays[:,dz]
        x   = rays[:,ox] + dxz*(z-rays[:,oz])
        dyz = rays[:,dy]/rays[:,dz]
        y   = rays[:,oy] + dyz*(z-rays[:,oz])        
        return x,y

    def make_histogram(self,x):
        xh=np.histogram(x,bins=np.arange(-self.lim-self.step,
                                         self.lim+self.step,
                                         self.step))
        return xh
        
        
    def trace(self,z,rays,rot=False, element=None):
        for n,z in enumerate(z):
            x,y      = self.trace_pos(rays,z)
            shiftx   = np.average(x)
            shifty   = np.average(y)
            y        = y-shifty
            x        = x-shiftx
            if n==0:
                xh         =  self.make_histogram(x)
                xh         =  xh[0]
                yh         =  self.make_histogram(y)
                yh         =  yh[0]
                argmax_0   =  np.argmax(yh)
                
            else:
                xh_temp    =  self.make_histogram(x)
                yh_temp    =  self.make_histogram(y)
                xh         =  np.concatenate((xh,xh_temp[0]), axis=0)
                yh         =  np.concatenate((yh,yh_temp[0]))
        xh=xh.reshape((n+1,xh_temp[0].shape[0]))  
        yh=yh.reshape((n+1,yh_temp[0].shape[0]))  

        xh=np.rot90(xh)
        if rot==False:
            return xh,np.rot90(yh)
        elif rot==True:
            return yh,np.rot90(np.flip(xh)) # no idea why I have to flip this
        
    def change_name(self, new_name, pos):
        self.element_names_list[pos] = new_name 
        
        
    def plot(self,save_img = True, img_name='test', extension='.png', show_img=False, annotate_OE=False,lim_top=False,lim_side=False, debug=False):
        dx, dy = self.step_z, self.step
        xmax = self.yh.shape[1]*self.step_z + dx
        y, x = np.mgrid[slice(-self.lim, self.lim + dy, dy),
            slice(0,xmax , dx)]
        x=x[:,:-1]
        y=y[:,:-1]
        if debug:
            print('######################')
            print('Dimension check')
            print('yh shape:  ', self.yh.shape)
            print('xh shape:  ', self.xh.shape)
            print('x shape:    ',x.shape)
            print('y shape:    ',y.shape)
            print('x :    ',x)
            print('self.yh.shape[1], self.step_z, dx', self.yh.shape, self.yh.shape[1], self.step_z, dx)
            print('self.yh.shape[1]*self.step_z + dx',self.yh.shape[1]*self.step_z + dx)
            print('DEBUG:: self.distances', self.distances)
            print('DEBUG:: self.distance_list', self.distance_list)
            print('DEBUG:: self.elements', self.elements)
            print('DEBUG:: self.elements_name_list', self.element_names_list)

        g=1.5
        plt.figure(self.count_fig)
        fig, (ax1, ax2) = plt.subplots(2,figsize=(6.4*g, 4.8*g))
        pcm=ax1.pcolormesh(x/1000,y,self.xh, cmap='inferno')
        ax1.clear()
        ax1.pcolormesh(x/1000,y,np.log(self.xh), cmap='inferno')
        ax2.pcolormesh(x/1000,y,np.log(self.yh), cmap='inferno')
        

        ax1.set_title('top view')
        ax2.set_title('side view')
        ax2.set_xlabel('[meters]')
        ax1.set_ylabel('Beam Width [mm]')
        ax2.set_ylabel('Beam Width [mm]')
        cbar1=fig.colorbar(pcm,ax=ax1) 
        cbar2=fig.colorbar(pcm,ax=ax2)
        cbar1.set_label('# of rays [a.u.]')#, rotation=-90) 
        cbar2.set_label('# of rays [a.u.]')
        posx=0
        xtick_pos = []
        xtick_label = []
        # put correct ticks and labels
        plot_length = xmax/1000
        beamline_length = sum(self.distance_list)/1000
        beamline_el_pos = np.cumsum(self.distance_list)/1000
        xtick_pos = (beamline_el_pos*plot_length)/beamline_length
        xtick_label = np.round(beamline_el_pos,2)

        # label the elements
        if annotate_OE == True:
            for ind, text in enumerate(self.element_names_list):
                if ind !=len(self.element_names_list):
                    posy =  -self.lim
                    posx = xtick_pos[ind]

                if ind<len(self.element_names_list)-1:
                    if (xtick_pos[ind+1]-xtick_pos[ind])<1:
                        posy += 3
                ax1.text(posx,posy,text, rotation=60)
                ax2.text(posx,posy,text, rotation=60)
        
        # remove xticks and labels which are too close to eaeachhc other
        for ind in range(xtick_pos.shape[0]-2, 1, -1):
            if ind>0:# < xtick_pos.shape[0]-1:
                if np.abs(xtick_pos[ind]-xtick_pos[ind+1])<1:
                    xtick_label = np.delete(xtick_label,ind)
                    xtick_pos = np.delete(xtick_pos,ind)

        if debug:
            print('DEBUG:: plot_length', plot_length)
            print('DEBUG:: self.distance_list', self.distance_list)
            print('DEBUG:: beamline_length', beamline_length)
            print('DEBUG:: beamline_el_pos', beamline_el_pos)
            print('DEBUG:: xtick_label', xtick_label)
            print('DEBUG:: xtick_pos', xtick_pos)
        ax1.set_xticks(xtick_pos)
        ax1.set_xticklabels(xtick_label, rotation = 45, ha="right")
        ax2.set_xticks(xtick_pos)
        ax2.set_xticklabels(xtick_label, rotation = 45, ha="right")

        if lim_top != False:
            ax1.set_xlim(lim_top[0],lim_top[1])
            ax1.set_ylim(lim_top[2],lim_top[3])
        if lim_side != False:
            ax2.set_xlim(lim_side[0],lim_side[1])
            ax2.set_ylim(lim_side[2],lim_side[3])
        if save_img == True:
            plt.savefig(os.path.join(self.directory,img_name+extension))
        plt.tight_layout()
        plt.close(0)
        if show_img == True:
            plt.show()
        self.count_fig += 1
            
        return self.xh,self.yh,x/1000,y






import schwimmbad

class SequentialRunnerPool:
    """Most simplistic implementation of the sequential runner pool
    """
    def map(self,func, opts):
        for item in opts:
            func(item)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass#return Tr

def RunPool(multiprocessing=True):
    """_summary_

    Args:
        multiprocessing (bool or integer, optional): 
        if set to False will use single process sequential runner. 
        if set to True will use multiprocessing code with number of cpu automatically detected
        if set to number will use multiprocessing with number of cpus equal to that number
        Defaults to True.


    Returns:
        _type_: _description_
    """
    if multiprocessing==False:
        return SequentialRunnerPool()
    elif multiprocessing==True:
        multiprocessing==None
    return schwimmbad.JoblibPool(multiprocessing)

# multiprocesing = True/False (autodetect CPU number)
# multiprocessing = 5 # use 5 CPUs


###############################################################################
# Reading/Writing/Modifying RML files
from . import xmltools as xml
from .xmltools import XmlAttributedNameElement,XmlElement

###############################################################################
class BeamlineElement(XmlAttributedNameElement):
    def __init__(self, name: str, attributes: dict, **kwargs):
        super().__init__("name",name, attributes, **kwargs)

###############################################################################
class ObjectElement(XmlAttributedNameElement):
    def __init__(self, name: str, attributes: dict, **kwargs):        
        super().__init__("id",name, attributes, **kwargs)


###############################################################################
class ParamElement(XmlElement):
    def __init__(self, name: str, attributes: dict, **kwargs):
        super().__init__(name, attributes, **kwargs)

    def __dir__(self):
        """enumerating child objects by its attibute name
        """
        children_names = [x._name for x in self._children]
        attr_name = list(self._attributes.keys())
        return children_names + attr_name + ['cdata']

    def __getattr__(self, key):
        matching_children = [x for x in self._children if x._name == key]
        if key in self._attributes:
            matching_children.append(self._attributes[key])
        if matching_children:
            if len(matching_children) == 1:
                self.__dict__[key] = matching_children[0]
                return matching_children[0]
            else:
                self.__dict__[key] = matching_children
                return matching_children
        else:
            raise AttributeError("'%s' has no attribute '%s'" % (self._name, key))


###############################################################################
class RMLFile:
    """Read/Write wrapper for the Ray RML files

    Args:
            filename (str, optional): path to rml file. Defaults to None.
            template (str, optional): path to rml file to use as template. 
                                      Defaults to None.
    """
    def __init__(self,filename:str=None,/,template:str=None) -> None:
        """
        Args:
            filename (str, optional): path to rml file. Defaults to None.
            template (str, optional): path to rml file to use as template. 
                                      Defaults to None.
        """        
        self._filename=filename
        self._template = template if template is not None else filename
        self.__known_classes = {"beamline":BeamlineElement, 
                                "object":ObjectElement,
                                "param":ParamElement}
        self.read()
        

    @property
    def template(self):
        return self._template

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self,value):
        self._filename = value


    ###################################
    def read(self,file:str=None):
        """Read rml file

        Args:
            file (str, optional): file name to read. If set to None will use template file name defined during initilizatino of the class. Defaults to None.
        """
        if file is None:
            file = self._template
        self._root = xml.parse(file,known_classes = self.__known_classes)      
        self.beamline = self._root.lab.beamline

    def xml(self):
        return xml.serialize(self._root).strip()

    def write(self,file:str=None):
        """Write the rml to :code:`file`

        Args:
            file (str, optional): filename . Defaults to None.
        """        
        if file is None:
            file = self._filename
        with open(file,"w") as f:
            f.write(self.xml())

    def __str__(self) -> str:
        return f"RMLFile('{self._filename}',template='{self._template}')"

    def __repr__(self) -> str:
        return f"RMLFile('{self._filename}',template='{self._template}')"

    
###############################################################################
# test function for the data parsing
def parse(filename:str):
    __known_classes = {"beamline":BeamlineElement, 
                        "object":ObjectElement,
                        "param":ParamElement}
    xml.parse(filename,known_classes = __known_classes)        
